/********************************** CF92.C *********************************

Copyright Symbios Logic Inc.


Author:		Erich Otto Symbios Logic Systems Engineer
Purpose:		53CF92 Demonstration Software

This software was written by Symbios Logic to develop and test new products.  It is not a 
general SCSI driver.  Symbios Logic assumes no liability for its use, misuse or inability to use.

Project files:
		CF92.C 		Main body of code
		TARGET92.C	Target Mode routines
		CF92ISR.C	Interrupt Service Routine
		CF92TEST.C	Automatic Test Module
		AT_DMA92.C	Hardware dependant DMA code (8237 compatible)
		AT_DMA92.H  DMA controller parameters
		CF92AT.H	Hardware addresses, global defines
		CF92EV.H	External variables


								  **********/

#include <stdio.h>
#include <dos.h>
#include "at_dma92.h"
#include "cf92at.h"
#include "cf92ev.h"


/****************** Initialize IBM microchannel DMA channel 3 ****************

initIBMdma(*start_addrs, numbytes, rw)

Input arguments:
start_addrs	= a pointer to the first element in the memory array which
		  will be accessed by the DMA controller
numbytes	= number of bytes to transfer
rw		= direction of DMA transfer: r (for read) or w (write).  Read
		  is an SCSI read, or Data In, which requires the DMA to write
		  to memory.  So, rw is backward for this module.

calls:		calc_abs_adrs(array_name) 	to calculate the absolute
						address of the memory array

!!! Machine Dependant Code !!!

This routine sets up the DMA controler on the AT (clone) motherboard to
transferdata between the 53CF92 and main memory via the motherboard DMA.

							   ***************/


void initIBMdma(void far *start_addrs, unsigned long numbytes, char rw) {
	calc_abs_adrs(start_addrs);			/* get array address for DMA controler */

	outportb(dma_request,ch3);			/* clear software request bit          */
	outportb(dma_mask,(0x40|ch3));		/* mask channel 3 while programming it */

	outportb(dma_flop,0);				/* reset byte pointer for memory address */
	outportb(dma3_addrs,lsB);			/* LSB of memory destination address     */
	outportb(dma3_addrs,mdB);			/* middle byte of 20-bit address         */
	outportb(dma3_page,msB);			/* MSB of memory destination address     */

	outportb(dma_flop,0);				/* reset byte pointer    */
	outportb(dma3_count,(numbytes) & 0xff);	/* LSB of transfer count */
	outportb(dma3_count,((numbytes) >> 8) & 0xff);	/* MSB of transfer count */

	if (parms.burst=='y')
		switch (rw)	{
			case 'r' :
				outportb(dma_mode,(demand|WT|ch3)); /* write demand mode */
				break;
			case 'w' :
				outportb(dma_mode,(demand|RD|ch3)); /* read demand mode */
				break;
			default  :
				outportb(dma_mode,(demand|RD|ch3)); /* read demand mode */
			}
	else
		switch (rw)	{
			case 'r' :
				outportb(dma_mode,(single|WT|ch3)); /* write single */
				break;
			case 'w' :
				outportb(dma_mode,(single|RD|ch3)); /* read single transfer */
				break;
			default  :
				outportb(dma_mode,(single|RD|ch3)); /* read single transfer */
			}

	outportb(dma_mask,ch3);			  	/* un-mask channel 3 */
	}

/********************** Calculate Absolute Address **************************

calc_abs_adrs(*ptr)

input argument:   ptr	= 32-bit pointer to the first element of the memory
			array which will be accessed by the DMA.  In C the
			array name IS the pointer to the first element, so
			the subscript is not necessary.

!!! Machine Dependant Code !!!

This routine translates Intel microprocessor segment:offset into a 20-bit
absolute address.  The external functions FP_SEG() and FP_OFF() are in the
Borland Turbo C Library; intended for use on PC, AT, and MicroChannel
architectures.  This routine will not be necessary on machines
(such as Motorala) that have linear memory addressing.

							   ***************/

void calc_abs_adrs (void far *ptr) {
	extern unsigned long seg, off, abs_adrs;

	seg = FP_SEG(ptr);					/* Far Pointer Segment */
	off = FP_OFF(ptr);					/* Far Pointer Offset */

	abs_adrs = ((seg << 4) + off);		/* 20-bit address */

	msB = abs_adrs >> 16;               /* Most Significant byte */
	mdB = abs_adrs >>  8;               /* Middle Byte */
	lsB = abs_adrs >>  0;               /* Least Significant Byte */
	}


