/********************************** CF92AT.H *********************************

Copyright Symbios Logic Inc.


Author:		Erich Otto Symbios Logic Systems Engineer
Purpose:		53CF92 Demonstration Software

This software was written by Symbios Logic to develop and test new products.  It is not a 
general SCSI driver.  Symbios Logic assumes no liability for its use, misuse or inability to use.

Project files:
		CF92.C 		Main body of code
		TARGET92.C	Target Mode routines
		CF92ISR.C	Interrupt Service Routine
		CF92TEST.C	Automatic Test Module
		AT_DMA92.C	Hardware dependant DMA code (8237 compatible)
		AT_DMA92.H  DMA controller parameters
		CF92AT.H	Hardware addresses, global defines
		CF92EV.H	External variables

				  **************/

#define true  1
#define false 0

#define base	0x100			/* base address of the 53CF92 */

#define tc_low		base+0		/* transfer counter lsb 	*/
#define tc_high		base+1		/* transfer counter msb		*/
#define fifo		base+2		/* fifo register		*/
#define cmd_reg		base+3		/* command register		*/
#define stat_reg	base+4		/* status register		*/
#define dest_ID		base+4		/* selection/reselection bus id */
#define int_reg		base+5		/* interrupt status register    */
#define srtimout	base+5		/* select/reselect timeout reg  */
#define seq_reg		base+6		/* sequence step register	*/
#define syncprd		base+6		/* synchronous transfer period  */
#define flags_reg	base+7		/* indicates # of bytes in fifo */
#define syncoff		base+7		/* synchronous offset register  */
#define config1		base+8		/* configuration register	*/
#define clkconv		base+9		/* clock conversion reg		*/
#define testreg		base+10		/* test mode register		*/
#define config2		base+11		/* Configuration 2 Register */
#define config3		base+12		/* Configuration 3 Register	*/
#define config4		base+13		/* Configuration 4 Register */
#define e_reg		base+14		/* 24-bit transfer count, RCB   */
#define FIFO_bottom	base+15		/* Reserve FIFO byte register   */


/* C90 family commands */

#define Clear_ACK		0x12
#define clr_fifo		0x01
#define DMA_Transfer_Info	0x90
#define nop			0x80
#define Receive_msg		0x28
#define Send_Message		0x20
#define set_atn			0x1a
#define soft_reset		0x02
#define scsi_reset		0x03
#define Transfer_Info		0x10


/* SCSI defines */

#define message_reject		0x07

/************************** Function Declarations **************************/

/* defined in CF92.C */
void main(void);
void init(void);
void load_regs(void);
void fillbuff(unsigned char pattern, unsigned int start);
void dumpbuf(void);
void FIFO_Test(void);
void x_count_test(void);
void setup(void);
void scsi_menu(void);
void tur (void);
void DATA_DRIVER(char rw, unsigned short cdb_len, int data_len);
void select(unsigned short cdb_len);
void dma_xfer_data(char far* start_addrs, int count, char rw);
void nonDMA_xfer_data(char far* start_addrs, int count, char rw);
void cmdcomp(void);
void message(void);
void DMA_CDB(char rw);
void rqsense(unsigned short cdb_len);
void Message_In(char rw);
void sync_nego(unsigned int period, unsigned int offset);
void read_regs(void);

/* defined in TARGET92.C */
void target_mode(void);
void Ttest_ready(void);
void Trqsense(void);
void Tread(void);
void Twrite(void);
void dis_resel(void);
void serv_ATN(void);

/* defined in CF92TEST.C */
void test_driver (void);
void asynch_test(void);
void synch_test(void);
void dma_cdb_test(void);
void check_buf(unsigned int start);
void EXSCSI(char rw);

/* defined in CF92ISR.C */
void isr(int phase);

/* defined in AT_DMA92.C */
void initIBMdma(void far *start_addrs, unsigned long numbytes, char rw);
void calc_abs_adrs (void far *ptr);
