/********************************** CF92ISR.C *********************************

Copyright Symbios Logic Inc.


Author:		Erich Otto Symbios Logic Systems Engineer
Purpose:		53CF92 Demonstration Software

This software was written by Symbios Logic to develop and test new products.  It is not a 
general SCSI driver.  Symbios Logic assumes no liability for its use, misuse or inability to use.

Project files:
		CF92.C 		Main body of code
		TARGET92.C	Target Mode routines
		CF92ISR.C	Interrupt Service Routine
		CF92TEST.C	Automatic Test Module
		AT_DMA92.C	Hardware dependant DMA code (8237 compatible)
		AT_DMA92.H  DMA controller parameters
		CF92AT.H	Hardware addresses, global defines
		CF92EV.H	External variables


				  **************/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "cf92at.h"
#include "cf92ev.h"


/********************** CF92 Interrupt Service Routine ***********************
isr(int phase)

Input argument = expected phase     0 = non-DMA Selection complete, Data Out
					1 = non-DMA Selection complete, Data In
					2 = DMA Transfer complete
					3 = Status good and Message good
					4 = expect Target to Disconnect
					5 = expect request for Status phase
					6 = Request Sense, variable data length
					7 = Non-Dma transfer complete
					8 = Message In Interrupt
					9 = Message Accepted Interrupt
				   10 = ReSelect
				   11 = Data In
				   12 = Data Out
				   13 = Sync Data in, odd byte
				   14 = DMA Selection complete, Data Out
				   15 = DMA Selection complete, Data In
				   16 = Select with ATN and Stop
				   17 = SDTR reply--bus service
				   18 = SDTR reply--funtion complete
				   19 = Test Unit Ready
				   20 = non-DMA Data Out intermediate
				   21 = non-DMA Data transfer complete

				   52 = Target non-DMA Data Transfer
				   53 = Target DMA Data Transfer Complete
				   54 = Target Disconnected Interrupt
				   55 = Target Send Message
				   56 = ReSelect Interrupt
				   57 = Target service ATN/Get message
				   58 = Target Reject Message

The interrupt service routine polls for an interrupt; reads the status,
interrupt status, and sequence step registers; dedices whether these values
are the expected ones; prints "completed" on the CRT if yes or "ERROR" and
a register dump if not.  The words "completed" or ERROR are printed on the
same line as the Phase description printed by one of the subroutine that
program the CF92.  The phase description is printed before CF92 is enabled,
and Completed or Error is printed after the interrupt is serviced.

							   ***************/

void isr(int phase) {
	switch (phase) {
		case 2: case 3: case 4:
			while (inportb(stat_reg) < 0x80);	/* Poll for interrupt */
			break;
		default:
			while (!kbhit() && (inportb(stat_reg) < 0x80)); /* Poll for interrupt */
			break;
		}

	status = (inportb(stat_reg) & 0x0ff);	/* CF92 status register */
	seq = (inportb(seq_reg) & 0x07);	/* sequence step register */
	intstat = inportb(int_reg);			/*interrupt status register*/
	fifo_flags = (inportb(flags_reg)&0x1f);		/* number bytes in FIFO	*/
	x_count = (inportb(tc_high) << 8) | inportb(tc_low);

	if(kbhit()) {
		temp=getch();					/* flush keyboard hit */

		printf("\n\nERROR!\n\nKeyboard hit while waiting for interrupt: isr(%d)",phase);
		printf("\n  Status Register    = %02X", status);
		if(status & 0x20)
			printf("  PARITY ERROR");
		printf("\n  Interrupt Register = %02X", intstat);
		printf("\n  Sequence Step Reg  = %02X", seq);
		printf("\n  FIFO flags         = %02X",fifo_flags);
		printf("\n  Transfer Count     = %04X",x_count);
		printf("\n  Command Register   = %02X",inportb(cmd_reg));
		printf("\n  Config 1 register  = %02X",inportb(config1));
		printf("\n  Config 2 register  = %02X",inportb(config2));
		printf("\n  Config 3 register  = %02X",inportb(config3));
		printf("\n  Config 4 register  = %02X",inportb(config4));

		if (fifo_flags > 0 )
			printf("\n  FIFO contents = ");
		for (i=0; i<fifo_flags; ++i)
			printf(" %02X",inportb(fifo));
		printf("\nCDB =");
		for (i = 0; i < 16; i++)
			printf (" %02X",cdb[i]);
		scsi_error = true;
		pause_flag = true;

		return;
		}

	switch (phase) {
		case 0:							/* Selection Complete, Data Out	*/
			if (status==0x80 && intstat==0x18 && seq==0x04)	{
				if(parms.detail == 'y') printf("complete");
				}
			else if (intstat==0x20) {
				printf("Timeout\n\n\n\n");
				scsi_error = true;
				}
			else if (status==0x87 && intstat== 0x18 && seq==0x04) {
				if(parms.detail == 'y') printf("complete");
				}
			else {
				printf("ERROR isr(0)\n");
				printf("\n  Status Register    = %02X   expected 80h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 18h", intstat);
				printf("\n  Sequence Step Reg  = %02X   expected 04h", seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0)
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; i++)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 1:							/* Selection Complete, Data In */
			if (status==0x81 && intstat==0x18 && seq == 4) {
				if(parms.detail == 'y') printf("complete");
				}
			else if (intstat==0x20) {
				printf("Timeout\n\n\n\n");
				scsi_error = true;
				}
			else if (status==0x87 && intstat==0x18 && seq==4) {
				if(parms.detail == 'y') printf("complete");
				}						/* Unexpected Message In */
			else {
				printf("ERROR isr(1)\n");
				printf("\n  Status Register    = %02X   expected 81h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 18h", intstat);
				printf("\n  Sequence Step Reg  = %02X   expected 04h", seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i=0; i<fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 2:							/* Data Transfer Complete Interrupt */
			if (status==0x93 && intstat==0x10) {
				if (parms.detail == 'y')
					printf("complete");
				outportb(cmd_reg,0x80);	/* clear Transfer Count Zero bit */
				}
			else {
				printf("ERROR isr(2)\n");
				printf("\n  Status Register    = %02X   expected 93h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 3: 						/* Status and Message Bytes accepted interrupt */
			if (fifo_flags == 2) {		/* get Status and Message bytes */
				statbyte = inportb(fifo);
				msg = inportb(fifo);
				}
			else {
				statbyte = 0xff;
				msg = 0xff;
				}

			if (status == 0x87 && intstat == 0x8 && statbyte == 0 && msg == 0) {
				if (parms.detail == 'y') printf("good");
				}
			else if (status == 0x87 && intstat == 0x08 && statbyte == 2 && msg == 0)
				printf("status = 02 CHECK CONDITION");
			else {
				printf("ERROR isr(3)\n");
				printf("\n  Status Register    = %02X   expected 87h",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 08h",intstat);
				printf("\n  Status Byte        = %02X   expected 00",statbyte);
				printf("\n  Message Byte       = %02X   expected 00",msg);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 4:  						/* Disconnected Interrupt */
			if(status==0x80 && intstat==0x20) {
				if (parms.detail == 'y')
					printf("\nDisconnected\n");
				}
			else {
				printf("\nMessage Accepted Command ERROR isr(4)\n");
				printf("\n  Status Register    = %02X   expected 80h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 20h", intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 5:  						/* Disconnected Interrupt */
			if(status==0x83 && intstat==0x18) {
				if(parms.detail == 'y')
					printf("complete");
				}
			else if (intstat==0x20)	{
				printf("Timeout\n\n\n\n");
				scsi_error = true;
				}
			else {
				printf("ERROR isr(5)\n");
				printf("\n  Status Register    = %02X   expected 83h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 18h", intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 6:							/* Request Sense, variable data block size, Data In */
			if(status==0x93 && intstat==0x10) {
				if (parms.detail == 'y')
					printf("complete");
				outportb(cmd_reg,0x80);	/* clear Transfer Count Zero bit */
				}
			else {                  	/* 8'th byte of data tells how    */
				/* many more data bytes to follow */
				printf("extended reply");
				dma_xfer_data (&dmabyte[8], (dmabyte[7] & 0x00ff), 'r');
				isr(2);
				}
			break;
		case 7  :						/* Bus Service, Data Phase */
			if (status & 0x20)
				printf("Parity ERROR isr(7)");
			if (((status & 0xee) != 0x80) || ((intstat & 0x10) != 0x10)) {
				printf("ERROR isr(7)\n");
				printf("\n  Status Register    = %02X   expected 100X 000X b",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected XXX1 XXXX b", intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 8 :						/* Message In phase */
			if(status==0x87 && intstat==0x08) {
				if(parms.detail == 'y')
					printf("complete");
				temp1 = inportb(fifo);	/* Message byte */
				}
			else {
				printf("ERROR ! isr(8)");
				printf("\n  Status Register    = %02X    expected 87 H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 08 H",intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 9 :						/* Message In phase */
			if(status==0x87 && intstat==0x10) {
				if (parms.detail == 'y') printf("message accepted");
				}
			else {
				printf("ERROR ! isr(9)");
				printf("\n  Status Register    = %02X    expected 87 H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 10 H",intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Transfer Count    = %04X",x_count);
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 10:						/* ReSelect */
			temp1=inportb(fifo);		/* Bus ID */
			temp2=inportb(fifo);		/* ID Message byte */

			if (status==0x87 && intstat==0x0c && (temp2 & 0x80)) {
				outportb(cmd_reg,0x12);	/* release SCSI ACK */
				if (parms.detail == 'y') {
					printf("\nReSelected by device %02X  ID Message = %02X",temp1,temp2);
					}
				}
			else {
				printf("ERROR isr(10)\n");
				printf("\n  Status Register    = %02X   expected 87h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 0Ch", intstat);
				printf("\n  Sequence Step Reg  = %02X", seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 11:						/* Data In, after ReSelection */
			if (status==0x81 && intstat==0x10) {
				if (parms.detail == 'y')
					printf("\nfunction complete");
				}
			else {
				printf("ERROR isr(11)\n");
				printf("\n  Status Register    = %02X   expected 81h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  Sequence Step Reg  = %02X", seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 12:						/* Data Out, after ReSelection */
			if (status==0x80 && intstat==0x10) {
				if (parms.detail == 'y')
					printf("\nfunction complete");
				}
			else {
				printf("ERROR isr(12)\n");
				printf("\n  Status Register    = %02X   expected 80h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  Sequence Step Reg  = %02X", seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 13 :						/* Synchronous Data In odd-byte */
			if(status==0x83 && intstat==0x10 && fifo_flags==1 && x_count==1) {
				printf("complete");
				}
			else {
				printf("ERROR isr(13)\n");
				printf("\n  Status Register    = %02X   expected   83h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected   10h", intstat);
				printf("\n  FIFO flags         = %02X   expected    1",fifo_flags);
				printf("\n  Transfer Count     = %04X expected 0001",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));
				printf("\n  FIFO contents      = ");

				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));
				printf("\n");
				scsi_error = true;
				}
			break;
		case 14:						/* DMA Selection Complete, Data Out */
			if (status==0x90 && intstat==0x18 && seq==0x04 && x_count == 0) {
				if(parms.detail == 'y')
					printf("complete");
				}
			else if (intstat==0x20) {
				printf("Timeout\n\n\n\n");
				scsi_error = true;
				}
			else if (status==0x97 && intstat== 0x18 && seq==0x04) {
				if(parms.detail == 'y')
					printf("complete");	/* unexpected Message In */
				}
			else {
				printf("ERROR isr(14)\n");
				printf("\n  Status Register    = %02X   expected 90h", status);
				if(status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 18h", intstat);
				printf("\n  Sequence Step Reg  = %02X   expected 04h", seq);
				printf("\n  FIFO flags         = %02X   expected  0",fifo_flags);
				printf("\n  Transfer Count     = %04X   expected 0",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 15:						/* Selection Complete, Data In */
			if (status==0x91 && intstat==0x18 && seq==0x04) {
				if(parms.detail == 'y')
					printf("complete");
				}
			else if (intstat==0x20) {
				printf("Timeout\n\n\n\n");
				scsi_error = true;
				}
			else if (status==0x97 && intstat== 0x18 && seq==0x04) {
				if (parms.detail == 'y')
					printf("complete");	/* Unexpected Message In */
				}
			else {
				printf("ERROR isr(15)\n");
				printf("\n  Status Register    = %02X   expected 91h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 18h", intstat);
				printf("\n  Sequence Step Reg  = %02X   expected 04h", seq);
				printf("\n  FIFO flags         = %02X   expected  0",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));
				if (fifo_flags > 0 )
					printf("  FIFO contents      = ");
				for (i=0; i<fifo_flags; ++i)
					printf("%02X ",inportb(fifo));
				printf("\n");

				scsi_error = true;
				}
			break;
		case 16:						/* Select with ATN and stop */
			if (status==0x86 && intstat==0x18 && seq==0x01) {
				if(parms.detail == 'y')
					printf("complete");
				}
			else if (intstat==0x20) {
				printf("Timeout, cannot select device\n\n");
				scsi_error = true;
				}
			else {
				printf("ERROR isr(16)\n");
				printf("\n  Status Register    = %02X   expected 86h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 18h", intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 17:						/* get targets SDTR reply */
			if (status == 0x87 && intstat == 0x10 && fifo_flags == 0) {
				if (parms.detail == 'y')
					printf("sent");
				}
			else {
				printf("ERROR isr(17)\n");
				printf("\n  Status Register    = %02X   expected 87h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  FIFO flags         = %02X   expected  0",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 18:						/* get targets SDTR reply */
			if (status==0x87 && intstat==0x08)
				return;
			else {
				printf("ERROR isr(18)\n");
				printf("  Status Register    = %02X   expected 87h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 08h", intstat);
				printf("\n  FIFO flags         = %02X   expected  0",fifo_flags);
				printf("\n  Transfer Count   = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0)
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 19:						/* Test Unit Ready */
			if (status==0x83 && intstat==0x10) {
				if (parms.detail == 'y')
					printf("complete");
				}
			else {
				printf("ERROR isr(%d)",phase);
				printf("\n  Status Register    = %02X   expected 83h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count   = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 20:						/* non-DMA Data Out */
			if (status==0x80 && intstat==0x10 && fifo_flags==0)
				return;
			else {
				printf("ERROR isr(%d)",phase);
				printf("\n  Status Register    = %02X   expected 80h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  FIFO flags         = %02X   expected 00 ",fifo_flags);
				printf("\n  Transfer Count   = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 21:						/* non-DMA Data Transfer Complete */
			if (status==0x83 && intstat==0x10) {
				if (parms.detail == 'y')
					printf("complete");
				}
			else {
				printf("ERROR isr(%d)",phase);
				printf("\n  Status Register    = %02X   expected 83h", status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X   expected 10h", intstat);
				printf("\n  FIFO flags         = %02X");
				printf("\n  Transfer Count   = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 52 :						/* Target Data Transfer Complete */
			if(intstat==0x08 && (status==0x90 || status==0x91))	{
				}
			else {
				printf("ERROR ! isr(%d)",phase);
				printf("\n  Status Register    = %02X    expected 90 H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 08 H",intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count   = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 53 :						/* Target Data Transfer Complete */
			if ((status & 0xf8)==0x90 && intstat==0x08)	{
				}
			else {
				printf("ERROR ! isr(%d)",phase);
				printf("\n  Status Register    = %02X    expected 90 H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 08 H",intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count   = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 54 :						/* Target Disconnected Interrupt */
			if (intstat == 0x28 && seq == 0x02)	{
				if (parms.detail == 'y')
					printf("good\nDisconnected");
				}
			else {
				printf("ERROR !  isr(%d)",phase);
				printf("\n  Status Register    = %02X",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 28 H",intstat);
				printf("\n  Sequence Step      = %02X    expected 02 H",seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 55 :						/* Target Data Transfer Complete */
			if ((status & 0xef) == 0x87 && intstat == 0x08)	{
				if (parms.detail == 'y')
					printf("complete");
				}
			else {
				printf("ERROR ! isr(%d)",phase);
				printf("\n  Status Register    = %02X    expected 87 or 97 H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 08 H",intstat);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count     = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0)
					printf("\n  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 56 :						/* Reselection Interrupt */
			if (intstat==0x08) {
				if (parms.detail == 'y')
					printf("complete");
				}
			else if (intstat==0x20)	{
				printf("  Time Out");
				scsi_error = true;
				}
			else {
				printf("ERROR!  isr(%d)",phase);
				printf("\n  Status Register    = %02X    expected 97 or 87 H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 08 H",intstat);
				printf("\n  Sequence Step      = %02X",seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count    = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		case 57 :						/* Target Service ATN/Get message */
			if ((status == 0x86 || status == 0x96) && (intstat == 0x18 || intstat == 0x08))
				return; 				/* good */
			else {
				printf("ERROR!  isr(%d)",phase);
				printf("\n  Status Register    = %02X    expected 86H",status);
				if (status & 0x20)
					printf("  PARITY ERROR");
				printf("\n  Interrupt Register = %02X    expected 18H or 08H",intstat);
				printf("\n  Sequence Step      = %02X",seq);
				printf("\n  FIFO flags         = %02X",fifo_flags);
				printf("\n  Transfer Count    = %04X",x_count);
				printf("\n  Command Register   = %02X",inportb(cmd_reg));
				printf("\n  Config 1 register  = %02X",inportb(config1));
				printf("\n  Config 2 register  = %02X",inportb(config2));
				printf("\n  Config 3 register  = %02X",inportb(config3));
				printf("\n  Config 4 register  = %02X",inportb(config4));

				if (fifo_flags > 0 )
					printf("  FIFO contents = ");
				for (i = 0; i < fifo_flags; ++i)
					printf("%02X ",inportb(fifo));

				scsi_error = true;
				}
			break;
		default:
			printf("\nInput argument error to isr().  phase = %d",phase);
			scsi_error = true;
		}								/* end of switch */
	}									/* end of isr  */
