/********************************** TARGET92.C *********************************

Copyright Symbios Logic Inc.


Author:		Erich Otto Symbios Logic Systems Engineer
Purpose:		53CF92 Demonstration Software

This software was written by Symbios Logic to develop and test new products.  It is not a 
general SCSI driver.  Symbios Logic assumes no liability for its use, misuse or inability to use.

Project files:
		CF92.C 		Main body of code
		TARGET92.C	Target Mode routines
		CF92ISR.C	Interrupt Service Routine
		CF92TEST.C	Automatic Test Module
		AT_DMA92.C	Hardware dependant DMA code (8237 compatible)
		AT_DMA92.H  DMA controller parameters
		CF92AT.H	Hardware addresses, global defines
		CF92EV.H	External variables



***************************** Target Mode Code ****************************/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "CF92AT.H"
#include "CF92EV.H"

void target_mode(void) {
	clrscr();
	printf("\n\n     Target Mode\nhit any key to Exit\n\n");

	do {
		outportb(cmd_reg,0x44);			/* enable Selection-Reselection */
		while (!kbhit() && (inportb(stat_reg) < 0x80)); /* Poll for interrupt */

		if (kbhit()) {
			temp = getch();				/* flush keyboard char */
			return;
			}

		status  = inportb(stat_reg);
		seq     = inportb(seq_reg) & 7;
		intstat = inportb(int_reg);
		fifo_flags = inportb(flags_reg) & 0x1f;

		if (intstat & 0x80)	{			/* SCSI Reset Detected */
			printf("\n SCSI Reset Detected");
			load_regs();
			}
		else if (intstat & 7) {			/* Selection or Reselection */
			if (parms.detail == 'y') {
				if (intstat & 0x01)
					printf("\n\nSelected");
				else if(intstat & 0x02)
					printf("\n\nSelected with ATN");
				else if(intstat & 0x04)
					printf("\n\nReselected");
				}

			if ((status & 7)==6 && intstat==0x12) {		/* Message Out */
				serv_ATN();				/* service ATN */
				if (scsi_error)
					return;
				goto itterate_do;
				}
			else if (!(status & 0x08)) {
				printf("\n  group 2, 3 or 4 Reserved Command ! ");
				for (i = 0; i < fifo_flags; ++i)
					printf(" %02X",inportb(fifo));
				scsi_error = true;
				getch();
				return;
				}

			bus_ID = inportb(fifo);		/* get bus ID */
			ID_msg = inportb(fifo);		/* get ID Message */

			fifo_flags = inportb(flags_reg) & 0x1f;
			for (i=0; i<fifo_flags; ++i) {	/* get CDB */
				cdb[i]=inportb(fifo);
				}

			if (cdb[5] == 0xAA) {
				printf("\n\nDEC Testing\n\n");
				parms.blkcount = cdb[4] & 0x0ff;

				if (ID_msg & 0x40)
					dis_resel();
				if (scsi_error)
					return;

				outportb(fifo,0x00);
				outportb(cmd_reg,0x22);

				for (i=0; i<185; ++i) {
					}
				outportb(fifo,0x00);
				outportb(cmd_reg,0x22);

				for (i=0; i<20; ++i) {
					}
				outportb(fifo,0x00);
				outportb(cmd_reg,0x22);

				while (!kbhit());
				}
			else {
				group_code = (cdb[0] & 0xe0) >> 5;
				switch (cdb[0])	{       /* 3rd byte = op-code */
					case 0x00:
						Ttest_ready();  /* Test Unit Ready */
						break;
					case 0x03:
						Trqsense();     /* Request Sense */
						break;
					case 0x08:
						parms.blkcount = cdb[4] & 0x0ff;
						Tread();        /* Data In */
						break;
					case 0x0a:
						parms.blkcount = cdb[4];
						Twrite();	  	/* Data Out */
						break;
					case 0x28:
						parms.blkcount = cdb[7] | (cdb[8]  << 8);
						Tread();        /* Data In */
						break;
					case 0x2a:
						parms.blkcount = cdb[7] | (cdb[8] << 8);
						Twrite();	  	/* Data Out */
						break;
					default:
						printf("\n unsupported command");
						for (i = 0; i < fifo_flags; ++i)
							printf("%02X ",cdb[i]);
					}					/* end of switch (cdb[0]) */
				}
			if (scsi_error)
				return;

			outportb(cmd_reg, clr_fifo);
			outportb(fifo,statbyte);    /* status byte defined in Tsubroutine */
			outportb(fifo,0);           /* no Message */
			if (parms.detail == 'y')
				printf("\nStatus and Message ... ");
			outportb(cmd_reg,0x24);     /* Terminate Sequence Command */
			isr(54);                    /* expect Disconnected interrupt */
			if (scsi_error)
				return;

			++ iocount;
			printf("%cnumber IO's = %8lX h    %12lu d",13,iocount,iocount);
			}							/* end of if(instat & 3) */
		else {
			printf("\nInterrupt detected, but not Selection or Reselection");
			printf("\n  Status Register    = %02X",status);
			printf("\n  Interrupt Register = %02X",intstat);
			printf("\n  Sequence Step      = %02X",seq);
			printf("\n  FIFO Flags         = %02X",fifo_flags);
			printf("\n  Command Register   = %02X",inportb(cmd_reg));
			scsi_error = true;
			return;
			}

		itterate_do:;
		} while (!kbhit());				/* loop forever */
	}               					/* end of target_mode */

/************************ Target Test Unit Ready **************************/

void Ttest_ready(void) {
	printf("\nTest Unit Ready");
	outportb(cmd_reg, clr_fifo);
	statbyte=0;							/* ready code */
	}

/************************** Target Request Sense ***************************/

void Trqsense(void) {
	outportb(cmd_reg, clr_fifo);
	outportb(fifo,0x70);   				/* This reply to Request Sense	*/
	outportb(fifo,0x00);
	outportb(fifo,0x05);
	outportb(fifo,0x00);
	outportb(fifo,0x00);
	outportb(fifo,0x41);
	outportb(fifo,0xff);
	outportb(fifo,0x00);

	if (parms.detail == 'y')
		printf("\nRequest Sense Data In ... ");
	outportb(cmd_reg,0x22);				/* non-DMA Send Data */
	isr(52);							/* non-DMA send data interrupt */

	if (scsi_error)
		statbyte = 2;					/* OK code */
	else {
		statbyte = 0;

		if(parms.detail == 'y')
			printf("complete");
		}
	}

/****************************** Target Data In *****************************/

void Tread(void) {
	unsigned int out_count, fifo_count;

	out_count = 0;

	if (ID_msg & 0x40)
		dis_resel();					/* disconnect privlidge? */
	if (scsi_error)
		return;

	if (parms.detail == 'y')
		printf("\nData In ... ");

	byte_count = ((long) parms.blksize) * parms.blkcount;

	if ((parms.DMA == 'y') || parms.sync_offset) {
		initIBMdma(dmabyte, byte_count, 'w');

		outportb(cmd_reg, clr_fifo);

		if (byte_count == 0x10000) {
			outportb(tc_low,0x00);
			outportb(tc_high,0x00);
			}
		else {
			outportb(tc_low,(byte_count & 0xff));
			outportb(tc_high,((byte_count >> 8) & 0xff));
			}
		outportb(cmd_reg, nop);
		outportb(cmd_reg,0xa2);			/* DMA send data */

		isr(53);						/* expect Data Transfer Complete */
		}
	else {								/* non-DMA transfers */
		do {
			if (byte_count - out_count > 16)	{	/* more than 16 bytes to send */
				for (i=0; i<16; ++i) {	/* fill FIFO */
					outportb(fifo, dmabyte[out_count]);
					++out_count;
					}
				outportb (cmd_reg, 0x22);	/* non-DMA send data */
				isr(52);
				if (scsi_error)
					break;
				}
			else {						/* 16 or fewer bytes to send */
				fifo_count = byte_count - out_count;
				for (i = 0; i < fifo_count; ++i) {	/* fill FIFO */
					outportb(fifo, dmabyte[out_count]);
					++out_count;
					}

				outportb (cmd_reg, 0x22);

				isr(52);    			/* expect data out phase */
				if (scsi_error)
					break;
				}
			} while (out_count < byte_count);
		}

	if (scsi_error) {
		statbyte = 2;
		return;
		}

	if(parms.detail == 'y')
		printf("complete");
	statbyte = 0;
	}

/****************************** Target Data Out ****************************/

void Twrite(void) {
	unsigned int out_count, fifo_count;

	out_count = 0;

	if (ID_msg & 0x40)
		dis_resel();					/* disconnect privlidge? */
	if (scsi_error)
		return;

	if (parms.detail == 'y')
		printf("\nData Out ... ");

	byte_count = parms.blksize * parms.blkcount;

	if ((parms.DMA == 'y') || parms.sync_offset) {
		initIBMdma(dmabyte, byte_count, 'r');

		outportb(cmd_reg, clr_fifo);
		outportb(tc_low,(byte_count & 0xff));
		outportb(tc_high,((byte_count >> 8) & 0xff));
		outportb(cmd_reg,0xaa);			/* DMA receive data */
		isr(53);						/* expect Data Transfer Complete */
		}
	else {								/* non-DMA transfers */
		while (out_count < byte_count) {
			outportb(cmd_reg, 0x2a);	/* non-DMA receive data */

			isr(52);    				/* expect data out phase */

			if (scsi_error)
				break;
			fifo_count = inportb(flags_reg) & 0x1F;
			for (i = 0; i < fifo_count; i++) {
				dmabyte[out_count] = inportb(fifo);
				out_count++;			/* increment counter */
				}
			}
		}
	if (scsi_error) {
		statbyte = 2;
		return;
		}

	if(parms.detail == 'y')
		printf("complete");
	statbyte = 0;
	}

/******************** Target Disconnect then ReSelect ********************/

void dis_resel(void) {
	if (parms.detail =='y') {
		printf("\nMessage In: Save Data pointers and Disconnect Message ... ");
		}
	outportb(cmd_reg, clr_fifo);
	outportb(fifo,2);					/* Save Data Pointers Message */
	outportb(fifo,4);					/* Disconnect message */
	outportb(cmd_reg,0x23);				/* Disconnect Sequence command */
	isr(54);
	if (scsi_error)
		return;

	for (i=800; i>0; ++i) {};			/* Time Delay before Reselect */

	if (parms.detail == 'y')
		printf("\n\nReSelect ... ");
	outportb(fifo, 0x80);				/* Identify message */
	outportb(cmd_reg, 0x40);			/* Reselect command */
	isr(56);
	}

/*************************** Target Service ATN **************************/

void serv_ATN(void) {
	int period, offset;

	if (parms.detail == 'y')
		printf("\nMessage Out ... ");
	do {								/* get message bytes ... */
		outportb(cmd_reg,Receive_msg);
		isr(57);
		} while (!scsi_error && intstat & 0x10);	/* while Bus Service */
	if (scsi_error)
		return;
	if (parms.detail == 'y')
		printf("complete");

	fifo_flags = inportb(flags_reg) & 0x1f;
	for (i = 0; i < fifo_flags; ++i)
		SDTR_in[i] = inportb(fifo);

	/*	FIFO 0	=	SDTR_in[0]	=	bus ID			*/
	/*	FIFO 1	=	SDTR_in[1]	=	ID message		*/
	/*	FIFO 2	=	SDTR_in[2]	=	Extended message	*/
	/*	FIFO 3	=	SDTR_in[3]	=	message length		*/
	/*	FIFO 4	=	SDTR_in[4]	=	SDTR message		*/
	/*	FIFO 5	=	SDTR_in[5]	=	encoded period		*/
	/*	FIFO 6	=	SDTR_in[6]	=	synchronous offset	*/

	if(SDTR_in[2] == 1 && SDTR_in[3] == 03 && SDTR_in[4] == 1) {
		period = SDTR_in[5] * 4;		/* convert message to nanoseconds */
		offset = SDTR_in[6];
		if (parms.detail == 'y') {
			printf("\n  SDTR Message received");
			for (i = 2; i < 7; ++i)
				printf(" %02X",SDTR_in[i]);
			printf("\n  Period = %d  Offset = %d",period,offset);
			}
		}
	else {
		printf("\n  ERROR! Unsupported Message ");
		for (i = 0; i < fifo_flags; ++i)
			printf("%02X ",SDTR_in[i]);
		scsi_error = true;
		return;
		}

	if (offset < 0 || offset > 16) {
		printf("\n  ERROR! Offset out of range (zero minimum 16 maximum)");
		/* in real SCSI, send message reject then default to  */
		/* asynchronous. But this is for testing ... so abort */
		scsi_error = true;
		return;
		}

	if (parms.FASTclk == 'y') {			/* if able to do FAST SCSI */
		if (period < 100 || period > 1000) {
			printf("\n  ERROR! Period out of range (200 ns minimum 1000 ns maximum)");
			/* in real SCSI, send message reject then default to  */
			/* asynchronous. But this is for testing ... so abort */
			scsi_error = true;
			return;
			}

		if (period < 200)
			parms.FAST_SCSI = 'y';

		period = period * 10;			/* gyrations for integer rounding */
		period = period / 25;
		period = period + 5;
		period = period / 10;
		parms.sync_period = period & 0x1f;
		}
	else {								/* else SCSI-1 */
		if (period < 200 || period > 1000) {
			printf("\n  ERROR! Period out of range (200 ns minimum 1000 ns maximum)\n");
			outportb(cmd_reg, clr_fifo);
			outportb(fifo, message_reject);
			outportb(cmd_reg, Send_Message);
			if (parms.detail == 'y')
				printf("\nReject SDTR Message ... ");
			isr(55);
			if (parms.detail=='y')
				printf("\nDisconnect\n");
			outportb(cmd_reg,0x27);		/* Disconnect */
			return;
			}
		parms.sync_period = (period / 40) & 0x1f;
		}

	if (parms.detail=='y')
		printf("\n  Synchronous Transfer Period register set to %d",parms.sync_period);
	if (parms.detail=='y')
		printf("\n  Synchronous Offset register set to %d",parms.sync_offset);
	if (parms.detail=='y')
		printf("\nMessage In; SDTR message ... ");

	for (i = 0; i < 4; ++i)
		SDTR_out[i] = SDTR_in[i+2];
	SDTR_out[4] = parms.sync_offset;

	for (i = 0; i < 5; ++i)
		outportb(fifo, SDTR_out[i]);

	outportb(cmd_reg,Send_Message);
	isr(55);
	if (scsi_error)
		return;

	if (parms.detail == 'y')
		printf("\nDisconnnect\n");
	outportb(cmd_reg,0x27);				/* Disconnect */
	load_regs();						/* load registers */
	}

